@@ -20,6 +20,7 @@ import re
 import sys
 import tempfile
 import time
+import urllib.parse
 from collections import namedtuple
 from contextlib import contextmanager
 from datetime import datetime
@@ -31,6 +32,9 @@ from packaging import version
 
 from ibm_watson_machine_learning import APIClient
 
+# WML Python API version with a fixed Assets.download function
+WML_HAS_FIXED_DOWNLOAD = "1000.0.0"
+
 LOGNAME = 'log.txt'
 
 
@@ -282,7 +286,7 @@ class DOWMLLib:
         self._logger.debug('Creating the WML client...')
         # http://ibm-wml-api-pyclient.mybluemix.net/#api-for-ibm-cloud
         client = APIClient(self._wml_credentials)
-        self._logger.info(f'Creating the client succeeded.  Client version is {client.version}')
+        self._logger.info(f'Creating the client succeeded.  Client version is {client.version}.')
         return client
 
     def _set_default_space(self):
@@ -330,9 +334,11 @@ class DOWMLLib:
     def client_data_asset_download(self, asset_id, filename):
         self._logger.debug(f'Downloading asset {asset_id} in {filename}...')
         with suppress_stdout():
-            # The return value is useless when filename is an absolute path
-            _ = self._client.data_assets.download(asset_id, filename)
-        self._logger.debug(f'Done.')
+            name = self._client.data_assets.download(asset_id, filename)
+            if version_is_greater(self._client.version, WML_HAS_FIXED_DOWNLOAD):
+                filename = name
+            # else the return value is useless when filename is an absolute path
+        self._logger.debug(f'Done saving {filename}.')
 
     def get_log(self, job_id):
         """Extract the engine log from the job.
@@ -342,12 +348,15 @@ class DOWMLLib:
         """
 
         def _get_asset_content(asset_id):
-            with tempfile.TemporaryDirectory() as temp_dir_name:
-                filename = os.path.join(temp_dir_name, f'{asset_id}-log.txt')
-                self.client_data_asset_download(asset_id, filename)
-                with open(filename) as f:
-                    content = f.read()
-                    return content
+            if version_is_greater(self._client.version, WML_HAS_FIXED_DOWNLOAD):
+                return self._client.data_assets.download(asset_id).decode('ascii')
+            else:
+                with tempfile.TemporaryDirectory() as temp_dir_name:
+                    filename = os.path.join(temp_dir_name, f'{asset_id}-log.txt')
+                    self.client_data_asset_download(asset_id, filename)
+                    with open(filename) as f:
+                        content = f.read()
+                        return content
 
         def _get_log_from_output_references(references):
             self._logger.debug(f'Looking for {LOGNAME} in output_data_references...')
@@ -355,14 +364,14 @@ class DOWMLLib:
                 if ref.get('type') != 'data_asset':
                     continue
                 if 'id' not in ref:
-                    self._logger.warning(f'Ignoring data asset with no id')
+                    self._logger.warning('Ignoring data asset with no id.')
                     continue
                 if ref['id'] == LOGNAME:
-                    self._logger.debug(f'Found it.')
+                    self._logger.debug('Found it.')
                     try:
                         asset_id = ref['location']['id']
                     except KeyError:
-                        self._logger.error('Log data asset has no location/id information')
+                        self._logger.error('Log data asset has no location/id information.')
                         break
                     self._logger.debug(f'This is asset {asset_id}.')
                     return _get_asset_content(asset_id)
@@ -373,7 +382,7 @@ class DOWMLLib:
             for output_data in outputs:
                 if output_data['id'] == LOGNAME:
                     if 'content' not in output_data:
-                        self._logger.error(f'Log without content for job {job_id}')
+                        self._logger.error(f'Log without content for job {job_id}.')
                         continue
                     self._logger.debug('Found it. Decoding it...')
                     output = output_data['content']
@@ -387,7 +396,7 @@ class DOWMLLib:
         try:
             do = job_details['entity']['decision_optimization']
         except KeyError:
-            self._logger.warning('No decision_optimization structure available for this job')
+            self._logger.warning('No decision_optimization structure available for this job.')
             return None
         # When we have references in the job, the 'output_data' may be an empty list
         if 'output_data' in do and do['output_data']:
@@ -395,37 +404,101 @@ class DOWMLLib:
         elif 'output_data_references' in do:
             return _get_log_from_output_references(do['output_data_references'])
         else:
-            self._logger.warning('No output_data or output_data_references structure available for this job')
+            self._logger.warning('No output_data or output_data_references structure available for this job.')
             return None
 
-    def get_output_assets(self, details):
-        """"Extract the output data asset ids from the job.
+    def _parse_asset_references(self, details, key):
+
+        def find_id_in_href(loc):
+            href = loc.get('href') if loc else None
+            if not href:
+                return None
+            path = urllib.parse.urlparse(href).path
+            s = re.search('/v2/assets/(.*)', path)
+            if s:
+                return s.group(1)
+            self._logger.warning(f'Could not decode href for asset \'{name}\'.')
+            return None
+
+        def find_id_in_id(loc):
+            return loc.get('id') if loc else None
 
-        :param details: The details of the job to get the output from
-        :return: A list of outputs. Each output is a tuple (name, id)
-        where the name is, well, the name of the output, and id is the
-        id of the corresponding data asset.
-        """
-        result = {}
         try:
-            refs = details['entity']['decision_optimization']['output_data_references']
+            refs = details['entity']['decision_optimization'][key]
         except KeyError:
-            self._logger.debug('No output references structure available for this job')
-            return result
+            self._logger.debug(f'No \'{key}\' structure available for this job.')
+            return {}
+        result = {}
         for ref in refs:
-            type = ref.get('type')
-            if type != 'data_asset':
-                self._logger.debug(f'Ignoring asset of unkown type \'{type}\'.')
+            asset_type = ref.get('type')
+            if asset_type != 'data_asset':
+                self._logger.debug(f'Ignoring asset of unknown type \'{asset_type}\'.')
+                continue
+            name = ref.get('id')
+            if not name:
+                self._logger.warning('Found a data asset with no name.')
                 continue
-            name = ref['id']
             self._logger.debug(f'Found a data asset named {name}.')
-            result[name] = ref['location']['id']
+            location = ref.get('location')
+            asset_id = find_id_in_id(location) or find_id_in_href(location)
+            if asset_id:
+                result[name] = asset_id
+            else:
+                self._logger.warning(f'Could not find asset id for asset \'{name}\'.')
         return result
 
-    def get_output(self, details, csv_as_dataframe=None, tabular_as_csv=False):
-        """"Extract the outputs from the job.
+    def get_output_asset_ids(self, details):
+        """"Extract the output data asset ids from the job details.
 
         :param details: The details of the job to get the output from
+        :return: A dict of outputs. Keys are the names of the outputs,
+        and the corresponding value for each key is the id of the asset.
+        """
+        return self._parse_asset_references(details, 'output_data_references')
+
+    def get_input_asset_ids(self, details):
+        """"Extract the input data asset ids from the job details.
+
+        :param details: The details of the job to get the output from
+        :return: A dict of inputs. Keys are the names of the inputs,
+        and the corresponding value for each key is the id of the asset.
+        """
+        return self._parse_asset_references(details, 'input_data_references')
+
+    def get_output(self, details, csv_as_dataframe=None, tabular_as_csv=False):
+        """Deprecated. Use get_outputs instead"""
+        return self.get_outputs(details, csv_as_dataframe, tabular_as_csv)
+
+    def _extract_inline_files_from_details(self, details, key, tabular_as_csv):
+        try:
+            files = details['entity']['decision_optimization'][key]
+        except KeyError:
+            self._logger.debug(f'No \'{key}\' structure available for this job.')
+            return {}
+        result = {}
+        for output_data in files:
+            name = output_data['id']
+            if 'content' in output_data:
+                # What we have here is a regular file, encoded
+                self._logger.debug(f'Found a regular file named {name}.')
+                content = self._extract_regular_file(output_data)
+                result[name] = content
+            elif ('values' in output_data and
+                  'fields' in output_data and
+                  name.lower().endswith('.csv')):
+                self._logger.debug(f'Found a CSV file named {name}.')
+                content = self._extract_csv_file(output_data, tabular_as_csv)
+                result[name] = content
+            else:
+                self._logger.warning(f'Found an unknown file named {name}.')
+                content = output_data
+                result[name] = content
+        return result
+
+    def get_outputs(self, details, csv_as_dataframe=None, tabular_as_csv=False):
+        """"Extract the inline outputs from the job details.
+
+        :param details: The details of the job to get the outputs from
         :param csv_as_dataframe: Whether the content of a CSV file should be
         returned as a Pandas DataFrame or not. Deprecated: use tabular_as_csv
         instead
@@ -442,30 +515,19 @@ class DOWMLLib:
             assert(tabular_as_csv is False)
             # Now we can replace it with the correct value
             tabular_as_csv = not csv_as_dataframe
-        result = {}
-        try:
-            outputs = details['entity']['decision_optimization']['output_data']
-        except KeyError:
-            self._logger.warning('No output structure available for this job')
-            return result
-        for output_data in outputs:
-            name = output_data['id']
-            if 'content' in output_data:
-                # What we have here is a regular file, encoded
-                self._logger.debug(f'Found a regular file named {name}')
-                content = self._extract_regular_file(output_data)
-                result[name] = content
-            elif ('values' in output_data and
-                  'fields' in output_data and
-                  name.lower().endswith('.csv')):
-                self._logger.debug(f'Found a CSV file named {name}')
-                content = self._extract_csv_file(output_data, tabular_as_csv)
-                result[name] = content
-            else:
-                self._logger.warning(f'Found an unknown file named {name}')
-                content = output_data
-                result[name] = content
-        return result
+        return self._extract_inline_files_from_details(details, 'output_data', tabular_as_csv)
+
+    def get_inputs(self, details, tabular_as_csv=False):
+        """"Extract the inline inputs from the job details.
+
+        :param details: The details of the job to get the inputs from
+        :param tabular_as_csv: Whether tabular inputs should be returned as
+        CSV file content instead of Pandas dataframe
+        :return: A dict of inputs, with the names of the assets as keys, and
+        the content as value, as bytes. We don't assume that the content is
+        actually text.
+        """
+        return self._extract_inline_files_from_details(details, 'input_data', tabular_as_csv)
 
     @staticmethod
     def _extract_csv_file(output_data, tabular_as_csv):
@@ -561,9 +623,9 @@ class DOWMLLib:
             if output.get('type') != 'data_asset':
                 continue
             if 'location' not in output:
-                self._logger.error(f'Missing \'location\' in details for job {job_id}')
+                self._logger.error(f'Missing \'location\' in details for job {job_id}.')
             elif 'id' not in output['location']:
-                self._logger.error(f'Missing \'location.id\' in details for job {job_id}')
+                self._logger.error(f'Missing \'location.id\' in details for job {job_id}.')
             else:
                 data_asset_id = output['location']['id']
                 self._logger.debug(f'Deleting data asset {data_asset_id}...')
@@ -571,7 +633,7 @@ class DOWMLLib:
                     self._client.data_assets.delete(data_asset_id)
                     self._logger.debug('Done.')
                 except WMLClientError:
-                    self._logger.error('Exception raised while trying to delete the asset', exc_info=True)
+                    self._logger.error('Exception raised while trying to delete the asset.', exc_info=True)
 
     def delete_job(self, job_id, hard=False):
         """Delete the given job.
@@ -648,35 +710,69 @@ class DOWMLLib:
                 names.append('Unknown')
         return names
 
-    class ProgressiveDelay:
-        def __init__(self):
-            self.delays = [2, 2, 2, 2, 2, 2, 3, 5, 5, 5, 5, 5, 5, 5, 5, 5,
-                           10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 30]
-
-        def wait(self):
-            delay = self.delays[0]
-            if len(self.delays) > 1:
-                self.delays.pop(0)
-            assert(2 <= delay <= 30)
-            time.sleep(delay)
-
     def wait_for_job_end(self, job_id, print_activity=False):
-        """Wait for the job to finish, return its status and details as a tuple."""
+        """Wait for the job to finish, return its status and details as a tuple.
+
+        If print_activity is True, some information is printed in the console."""
+
+        class ProgressiveDelay:
+            def __init__(self):
+                self.delays = [2, 2, 2, 2, 2, 2, 3, 5, 5, 5, 5, 5, 5, 5, 5, 5,
+                               10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 30]
+
+            def wait(self):
+                delay = self.delays[0]
+                if len(self.delays) > 1:
+                    self.delays.pop(0)
+                assert (2 <= delay <= 30)
+                time.sleep(delay)
+
+            def __enter__(self):
+                return self
+
+            def __exit__(self, exc_type, exc_val, exc_tb):
+                pass
+
+        class StatusLogger:
+            def __init__(self, initial_state):
+                self.last_state = initial_state
+                print(initial_state, end='', flush=True)
+
+            def log_state(self, state):
+                if state == self.last_state:
+                    print('.', end='', flush=True)
+                else:
+                    if self.last_state != '':
+                        print('')
+                    # else: if state was empty, no need to end the line
+                    print(state, end='', flush=True)
+                    self.last_state = state
+
+            def __enter__(self):
+                return self
+
+            def __exit__(self, exc_type, exc_val, exc_tb):
+                pass
+
         client = self._get_or_make_client()
-        delayer = DOWMLLib.ProgressiveDelay()
-        while True:
-            job_details = self.client_get_job_details(client, job_id, with_filter='solve_state,status')
-            do = job_details['entity']['decision_optimization']
-            status = self._get_job_status_from_details(job_details)
-            self._logger.info(f'Job status: {status}')
-            if status in ['completed', 'failed', 'canceled']:
-                break
-            else:
+        with StatusLogger('') as status_logger, \
+                ProgressiveDelay() as delayer:
+            while True:
+                job_details = self.client_get_job_details(client, job_id, with_filter='solve_state,status')
+                do = job_details['entity']['decision_optimization']
+                status = self._get_job_status_from_details(job_details)
+                self._logger.info(f'Job status: {status}')
+                if print_activity:
+                    status_logger.log_state(f'Job is {status}.')
+                if status in ['completed', 'failed', 'canceled']:
+                    break
                 if print_activity:
                     # There may be a bit of log to look at
                     try:
                         activity = do['solve_state']['latest_engine_activity']
                         if activity:
+                            # Because of the StatusLogger, we're not at the beginning of a line
+                            print('')
                             # We are joining the lines in the activity with a CR,
                             # only to remove them if they were already included...
                             # FIXME: what a waste!
@@ -686,7 +782,10 @@ class DOWMLLib:
                     except KeyError:
                         # This must mean that no activity is available yet
                         pass
-            delayer.wait()
+                delayer.wait()
+        if print_activity:
+            # The status_logger printed something, but didn't end the line yet
+            print('')
         return status, job_details
 
     @staticmethod
@@ -926,7 +1025,7 @@ class DOWMLLib:
         self._logger.debug('Done.')
         resources = details['models']['resources']
         model_name = f'{self.MODEL_NAME}-{self.model_type}-{self.do_version}'
-        self._logger.debug(f'Got the list. Looking for model named \'{model_name}\'')
+        self._logger.debug(f'Got the list. Looking for model named \'{model_name}\'...')
         model_id = None
         for r in resources:
             if r['metadata']['name'] == model_name:
@@ -991,7 +1090,7 @@ class DOWMLLib:
         self._logger.debug('Fetching existing spaces...')
         space_details = client.spaces.get_details()
         resources = space_details['resources']
-        self._logger.debug(f'Got the list. Looking for space named \'{self.space_name}\'')
+        self._logger.debug(f'Got the list. Looking for space named \'{self.space_name}\'...')
         space_id = None
         for r in resources:
             if r['entity']['name'] == self.space_name:
@@ -1083,5 +1182,5 @@ class DOWMLLib:
             if self.delete_asset(asset_to_delete):
                 self._logger.debug('Done.')
             else:
-                self._logger.warning('Could not delete pre-existing asset')
+                self._logger.warning('Could not delete pre-existing asset.')
         return data_asset_id
