@@ -63,7 +63,7 @@ job id, but none is specified, the last one is used.
 
 ''')
 
-    def __init__(self, wml_cred_file, space_id=None):
+    def __init__(self, wml_cred_file=None, space_id=None):
         super().__init__()
         self.lib = DOWMLLib(wml_cred_file, space_id)
         self.jobs = []
@@ -256,7 +256,8 @@ Wildcards (* and ?) are accepted (but no ~)."""
 Waits until the job is finished, printing activity. Hit Ctrl-C to interrupt.
 job is either a job number or a job id. Uses current job if not specified."""
         job_id = self._get_and_remember_job_id(job_id)
-        self.lib.wait_for_job_end(job_id, True)
+        status, _ = self.lib.wait_for_job_end(job_id, True)
+        print(f'Job has finished with status \'{status}\'.')
 
     def _cache_jobs(self):
         jobs = self.lib.get_jobs()
@@ -303,31 +304,44 @@ job is either a job number or a job id. Uses current job if not specified."""
             f.write(content)
 
     def do_output(self, job_id):
-        """output [job]
-Downloads all the outputs of a job, as well as the details/status of the job.
+        """Deprecated. Use 'dump' instead."""
+        self.do_dump(job_id)
+
+    def do_dump(self, job_id):
+        """dump [job]
+Downloads and stores all the inputs and outputs of a job, as well as the details/status of the job.
 job is either a job number or a job id. Uses current job if not specified."""
 
+        def store_inline_inputs(job_details):
+            inputs = self.lib.get_inputs(job_details, tabular_as_csv=True)
+            for name in inputs:
+                self.save_content(job_id, name, inputs[name])
+
         def store_inline_outputs(job_details):
-            outputs = self.lib.get_output(job_details, tabular_as_csv=True)
+            outputs = self.lib.get_outputs(job_details, tabular_as_csv=True)
             for name in outputs:
                 self.save_content(job_id, name, outputs[name])
 
-        def download_output_data_assets(job_details):
-            outputs = self.lib.get_output_assets(job_details)
-            for name in outputs:
+        def download_data_assets(assets):
+            for name in assets:
                 try:
                     os.mkdir(job_id)
                 except FileExistsError:
                     pass
-                asset_id = outputs[name]
+                asset_id = assets[name]
                 filename = f'{job_id}/{name}'
                 print(f'Storing {filename}')
                 self.lib.client_data_asset_download(asset_id, filename)
 
         job_id = self._get_and_remember_job_id(job_id)
         details = self.lib.get_job_details(job_id, with_contents='full')
+        store_inline_inputs(details)
         store_inline_outputs(details)
-        download_output_data_assets(details)
+        download_data_assets(self.lib.get_output_asset_ids(details))
+        # It's important that the input assets are stored last. This allows
+        # the user to cancel the transfer of these items, which probably would
+        # be the largest for a job, after the others have been stored.
+        download_data_assets(self.lib.get_input_asset_ids(details))
         # We don't want to store all the outputs in the details themselves
         self.lib.filter_large_chunks_from_details(details)
         details = pprint.pformat(details)
